# Copyright 2019 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#            http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
`compute.py`
Compute utilities.
"""

import logging
import time
import pprint
from slo_generator import utils
from slo_generator.report import SLOReport

LOGGER = logging.getLogger(__name__)


def compute(slo_config,
            error_budget_policy,
            timestamp=None,
            client=None,
            do_export=False,
            delete=False):
    """Run pipeline to compute SLO, Error Budget and Burn Rate, and export the
    results (if exporters are specified in the SLO config).

    Args:
        slo_config (dict): SLO configuration.
        error_budget_policy (dict): Error Budget policy configuration.
        timestamp (int, optional): UNIX timestamp. Defaults to now.
        client (obj, optional): Existing metrics backend client.
        do_export (bool, optional): Enable / Disable export. Default: False.
        delete (bool, optional): Enable / Disable delete mode. Default: False.
    """
    if timestamp is None:
        timestamp = time.time()

    # Compute SLO, Error Budget, Burn rates and make report
    exporters = slo_config.get('exporters')
    reports = []
    for step in error_budget_policy:
        report = SLOReport(config=slo_config,
                           step=step,
                           timestamp=timestamp,
                           client=client,
                           delete=delete)

        if report.is_empty():  # report is empty
            continue

        if delete:  # delete mode is enabled
            continue

        LOGGER.info(report)
        json_report = report.to_json()
        reports.append(json_report)

        if exporters is not None and do_export is True:
            export(json_report, exporters)

    return reports


def export(data, exporters):
    """Export data using selected exporters.

    Args:
        data (dict): Data to export.
        exporters (list): List of exporter configurations.

    Returns:
        obj: Return values from exporters output.
    """
    LOGGER.debug(f'Exporters: {pprint.pformat(exporters)}')
    LOGGER.debug(f'Data: {pprint.pformat(data)}')
    results = []

    # Passing one exporter as a dict will work for convenience
    if isinstance(exporters, dict):
        exporters = [exporters]

    for config in exporters:
        LOGGER.debug(f'Exporter config: {pprint.pformat(config)}')
        exporter_class = config.get('class')
        LOGGER.info(f'Exporting results to {exporter_class}')
        exporter = utils.get_exporter_cls(exporter_class)()
        ret = exporter.export(data, **config)
        results.append(ret)
        LOGGER.debug(f'Exporter return: {pprint.pformat(ret)}')



# Copyright 2019 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#            http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
`elasticsearch.py`
ElasticSearch backend implementation.
"""

import logging

from elasticsearch import Elasticsearch

LOGGER = logging.getLogger(__name__)

DEFAULT_DATE_FIELD = '@timestamp'


class ElasticsearchBackend:
    """Backend for querying metrics from ElasticSearch.

    Args:
        client (elasticsearch.ElasticSearch): Existing ES client.
        es_config (dict): ES client configuration.
    """

    def __init__(self, client=None, **es_config):
        self.client = client
        if self.client is None:
            self.client = Elasticsearch(**es_config)

    # pylint: disable=unused-argument
    def good_bad_ratio(self, timestamp, window, slo_config):
        """Query two timeseries, one containing 'good' events, one containing
        'bad' events.

        Args:
            timestamp (int): UNIX timestamp.
            window (int): Window size (in seconds).
            slo_config (dict): SLO configuration.

        Returns:
            tuple: A tuple (good_event_count, bad_event_count)
        """
        measurement = slo_config['backend']['measurement']
        index = measurement['index']
        query_good = measurement['query_good']
        query_bad = measurement.get('query_bad')
        query_valid = measurement.get('query_valid')
        date_field = measurement.get('date_field', DEFAULT_DATE_FIELD)

        # Build ELK request bodies
        good = ES.build_query(query_good, window, date_field)
        bad = ES.build_query(query_bad, window, date_field)
        valid = ES.build_query(query_valid, window, date_field)

        # Get good events count
        response = self.query(index, good)
        good_events_count = ES.count(response)

        # Get bad events count
        if query_bad is not None:
            response = self.query(index, bad)
            bad_events_count = ES.count(response)
        elif query_valid is not None:
            response = self.query(index, valid)
            bad_events_count = ES.count(response) - good_events_count
        else:
            raise Exception("`filter_bad` or `filter_valid` is required.")

        return (good_events_count, bad_events_count)

    def query(self, index, body):
        """Query ElasticSearch server.

        Args:
            index (str): Index to query.
            body (dict): Query body.

        Returns:
            dict: Response.
        """
        return self.client.search(index=index, body=body)

    @staticmethod
    def count(response):
        """Count event in Prometheus response.

        Args:
            response (dict): Prometheus query response.

        Returns:
            int: Event count.
        """
        try:
            return response['hits']['total']['value']
        except KeyError as exception:
            LOGGER.warning("Couldn't find any values in timeseries response")
            LOGGER.debug(exception)
            return 0

    @staticmethod
    def build_query(query, window, date_field=DEFAULT_DATE_FIELD):
        """Build ElasticSearch query.

        Add window to existing query.
        Replace window for different error budget steps on-the-fly.

        Args:
            body (dict): Existing query body.
            window (int): Window in seconds.
            date_field (str): Field to filter time on (must be an ElasticSearch
                field of type `date`. Defaults to `@timestamp` (Logstash-
                generated date field).

        Returns:
            dict: Query body with range clause added.
        """
        if query is None:
            return None
        body = {"query": {"bool": query}, "track_total_hits": True}
        range_query = {
            f"{date_field}": {
                "gte": f"now-{window}s/s",
                "lt": "now/s"
            }
        }

        # If a 'filter' clause already exist, add the range query on top,
        # otherwise create the 'filter' clause.
        if "filter" in body["query"]["bool"]:
            body["query"]["bool"]["filter"]["range"] = range_query
        else:
            body["query"]["bool"] = {"filter": {"range": range_query}}

        return body


ES = ElasticsearchBackend



# Copyright 2019 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#            http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
`cli.py`
Command-Line interface of `slo-generator`.
"""

import argparse
import logging
import sys

from slo_generator.compute import compute
import slo_generator.utils as utils

LOGGER = logging.getLogger(__name__)


def main():
    """slo-generator CLI entrypoint."""
    args = parse_args(sys.argv[1:])
    cli(args)


def cli(args):
    """Main CLI function.

    Args:
        args (Namespace): Argparsed CLI parameters.

    Returns:
        dict: Dict of all reports indexed by config file path.
    """
    utils.setup_logging()
    export = args.export
    delete = args.delete
    timestamp = args.timestamp

    # Load error budget policy
    LOGGER.debug(f"Loading Error Budget config from {args.error_budget_policy}")
    eb_path = utils.normalize(args.error_budget_policy)
    eb_policy = utils.parse_config(eb_path)

    # Parse SLO folder for configs
    slo_configs = utils.list_slo_configs(args.slo_config)
    if not slo_configs:
        LOGGER.error(f'No SLO configs found in SLO folder {args.slo_config}.')

    # Load SLO configs and compute SLO reports
    all_reports = {}
    for path in slo_configs:
        slo_config_name = path.split("/")[-1]
        LOGGER.debug(f'Loading SLO config "{slo_config_name}"')
        slo_config = utils.parse_config(path)
        reports = compute(slo_config,
                          eb_policy,
                          timestamp=timestamp,
                          do_export=export,
                          delete=delete)
        all_reports[path] = reports
    LOGGER.debug(all_reports)
    return all_reports


def parse_args(args):
    """Parse CLI arguments.

    Args:
        args (list): List of args passed from CLI.

    Returns:
        obj: Args parsed by ArgumentParser.
    """
    parser = argparse.ArgumentParser()
    parser.add_argument('--slo-config',
                        '-f',
                        type=str,
                        required=False,
                        help='SLO configuration file (JSON / YAML)')
    parser.add_argument('--error-budget-policy',
                        '-b',
                        type=str,
                        required=False,
                        default='error_budget_policy.yaml',
                        help='Error budget policy file (JSON / YAML)')
    parser.add_argument('--export',
                        '-e',
                        type=utils.str2bool,
                        nargs='?',
                        const=True,
                        default=False,
                        help="Export SLO reports")
    parser.add_argument('--delete',
                        '-d',
                        type=utils.str2bool,
                        nargs='?',
                        const=True,
                        default=False,
                        help="Delete SLO (use for backends with APIs).")
    parser.add_argument('--timestamp',
                        '-t',
                        type=int,
                        default=None,
                        help="End timestamp for query.")
    return parser.parse_args(args)


if __name__ == '__main__':
    main()



# Copyright 2019 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#            http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
`utils.py`
Utility functions.
"""
from datetime import datetime
import argparse
import collections
import glob
import importlib
import logging
import os
import pprint
import re
import sys
import warnings
import yaml

import pytz

from google.auth._default import _CLOUD_SDK_CREDENTIALS_WARNING

LOGGER = logging.getLogger(__name__)


def list_slo_configs(path):
    """List all SLO configs from path.

    If path is a file, normalize the path and return it as a list with one
    element.

    If path is a folder, get all SLO configs from folder (files starting with
    slo_*), normalize their paths and return them as a list.
    """
    path = normalize(path)
    if os.path.isfile(path):
        paths = [path]
    elif os.path.isdir(path):
        paths = sorted(glob.glob(f'{path}/slo_*.yaml'))
    else:
        raise Exception(f'SLO Config path "{path}" is not a file or folder.')
    return paths


def parse_config(path, ctx=os.environ):
    """Load a yaml configuration file and resolve environment variables in it.

    Args:
        path (str): the path to the yaml file.
        ctx (dict): Context to replace env variables from (defaults to
            `os.environ`).

    Returns:
        dict: Parsed YAML dictionary.
    """
    pattern = re.compile(r'.*?\${(\w+)}.*?')

    def replace_env_vars(content, ctx):
        """Replace env variables in content from context.

        Args:
            content (str): String to parse.
            ctx (dict): Context to replace vars from.

        Returns:
            str: the parsed string with the env var replaced.
        """
        match = pattern.findall(content)
        if match:
            full_value = content
            for var in match:
                try:
                    full_value = full_value.replace(f'${{{var}}}', ctx[var])
                except KeyError as exception:
                    LOGGER.error(f'Environment variable "{var}" should be set.')
                    raise exception
            content = full_value
        return content

    with open(path) as config:
        content = config.read()
        content = replace_env_vars(content, ctx)
        data = yaml.safe_load(content)

    LOGGER.debug(pprint.pformat(data))
    return data


def setup_logging():
    """Setup logging for the CLI."""
    debug = os.environ.get("DEBUG", "0")
    print("DEBUG: %s" % debug)
    if debug == "1":
        level = logging.DEBUG
    else:
        level = logging.INFO
    logging.basicConfig(stream=sys.stdout,
                        level=level,
                        format='%(name)s - %(levelname)s - %(message)s',
                        datefmt='%m/%d/%Y %I:%M:%S')
    logging.getLogger('googleapiclient').setLevel(logging.ERROR)

    # Ignore Cloud SDK warning when using a user instead of service account
    warnings.filterwarnings("ignore", message=_CLOUD_SDK_CREDENTIALS_WARNING)


def get_human_time(timestamp, timezone="Europe/Paris"):
    """Get human-readable timestamp from UNIX timestamp.

    Args:
        timestamp (int): UNIX timestamp.

    Returns:
        str: Formatted timestamp in ISO format.
    """
    date = datetime.fromtimestamp(timestamp, pytz.timezone(timezone))
    timeformat = '%Y-%m-%dT%H:%M:%S.%fZ'
    return datetime.strftime(date, timeformat)


def normalize(path):
    """Converts a path to an absolute path.

    Args:
        path (str): Input path.

    Returns:
        str: Absolute path.
    """
    return os.path.abspath(path)


def get_backend_cls(backend):
    """Get backend class.

    Args:
        backend (str): Exporter type.

    Returns:
        class: Backend class.
    """
    filename = re.sub(r'(?<!^)(?=[A-Z])', '_', backend).lower()
    return import_dynamic(f'slo_generator.backends.{filename}',
                          f'{backend}Backend',
                          prefix="backend")


def get_exporter_cls(exporter):
    """Get exporter class.

    Args:
        exporter (str): Backend type.

    Returns:
        class: Exporter class.
    """
    filename = re.sub(r'(?<!^)(?=[A-Z])', '_', exporter).lower()
    return import_dynamic(f'slo_generator.exporters.{filename}',
                          f'{exporter}Exporter',
                          prefix="exporter")


def import_dynamic(package, name, prefix="class"):
    """Import class or method dynamically from package and name.

    Args:
        package: Where the method or class is located in the import path.
        name: Name of method or class.

    Returns:
        obj: Imported class or method object.
    """
    try:
        return getattr(importlib.import_module(package), name)
    except Exception as exception:  # pylint: disable=W0703
        LOGGER.error(
            f'{prefix.capitalize()} "{package}.{name}" not found, check '
            f'package and class name are valid, or that importing it doesn\'t '
            f'result in an exception.')
        LOGGER.debug(exception)
        sys.exit(1)


def dict_snake_to_caml(data):
    """Convert dictionary with keys written in snake_case to another one with
    keys written in CamlCase.

    Args:
        data (dict): Input dictionary.

    Returns:
        dict: Output dictionary.
    """

    def snake_to_caml(word):
        return re.sub('_.', lambda x: x.group()[1].upper(), word)

    return apply_func_dict(data, snake_to_caml)


def apply_func_dict(data, func):
    """Apply function on a dictionary keys.

    Args:
        data (dict): Input dictionary.

    Returns:
        dict: Output dictionary.
    """
    if isinstance(data, collections.Mapping):
        return {func(k): apply_func_dict(v, func) for k, v in data.items()}
    return data


def str2bool(string):
    """Convert a string to a boolean.

    Args:
        string (str): String to convert

    Returns:
        bool: Boolean value.

    Raises:
        `argparse.ArgumentTypeError`: IF no acceptable boolean string is found.
    """
    if isinstance(string, bool):
        return string
    if string.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    if string.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    raise argparse.ArgumentTypeError('Boolean value expected.')



# Copyright 2019 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#            http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
`stackdriver.py`
Stackdriver Monitoring exporter class.
"""
import logging
from prometheus_client import CollectorRegistry, Gauge, push_to_gateway
from prometheus_client.exposition import basic_auth_handler, default_handler

LOGGER = logging.getLogger(__name__)
DEFAULT_METRIC_TYPE = "error_budget_burn_rate"
DEFAULT_METRIC_DESCRIPTION = ("Speed at which the error budget for a given"
                              "aggregation window is consumed")
DEFAULT_PUSHGATEWAY_URL = "http://localhost:9091"
DEFAULT_PUSHGATEWAY_JOB = "slo-generator"


class PrometheusExporter:
    """Prometheus exporter class."""

    def __init__(self):
        self.username = None
        self.password = None

    def export(self, data, **config):
        """Export data to Prometheus.

        Args:
            data (dict): Data to send to Prometheus.
            config (dict): Prometheus config.
                url (str, optional): Prometheus URL.
                custom_metric_type (str, optional): Custom metric type.
                custom_metric_unit (str, optional): Custom metric unit.
                custom_metric_description (str, optional): Custom metric
                    description.

        Returns:
            object: Prometheus API result.
        """
        self.create_timeseries(data, **config)

    def create_timeseries(self, data, **config):
        """Create Prometheus timeseries.

        Args:
            data (dict): Data to send to Prometheus.
            config (dict): Metric / exporter config.

        Returns:
            object: Metric descriptor.
        """
        metric_type = config.get('metric_type', DEFAULT_METRIC_TYPE)
        metric_description = config.get('metric_description',
                                        DEFAULT_METRIC_DESCRIPTION)
        prometheus_push_url = config.get('url', DEFAULT_PUSHGATEWAY_URL)
        prometheus_push_job_name = config.get('job', DEFAULT_PUSHGATEWAY_JOB)
        burn_rate = data['error_budget_burn_rate']

        # Write timeseries w/ metric labels.
        labels = {
            'service_name':
                data['service_name'],
            'feature_name':
                data['feature_name'],
            'slo_name':
                data['slo_name'],
            'window':
                str(data['window']),
            'error_budget_policy_step_name':
                str(data['error_budget_policy_step_name']),
            'alerting_burn_rate_threshold':
                str(data['alerting_burn_rate_threshold']),
        }
        registry = CollectorRegistry()
        gauge = Gauge(metric_type,
                      metric_description,
                      registry=registry,
                      labelnames=labels.keys())
        gauge.labels(*labels.values()).set(burn_rate)

        # Handle headers
        handler = default_handler
        if 'username' in config and 'password' in config:
            self.username = config['username']
            self.password = config['password']
            handler = PrometheusExporter.auth_handler

        return push_to_gateway(prometheus_push_url,
                               job=prometheus_push_job_name,
                               grouping_key=labels,
                               registry=registry,
                               handler=handler)

    def auth_handler(self, url, method, timeout, headers, data):
        """Handles authentication for pushing to Prometheus gateway.

        Args:
            url (str): Prometheus gateway URL.
            method (str): Prometheus query method.
            timeout (int): Prometheus timeout.
            headers (dict): Headers.
            data (dict): Data to send.

        Returns:
            func: Auth handler function.
        """
        username = self.username
        password = self.password
        return basic_auth_handler(url, method, timeout, headers, data, username,
                                  password)



# Copyright 2019 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#            http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
`pubsub.py`
Pubsub exporter class.
"""
import json
from google.cloud import pubsub_v1


class PubsubExporter:  # pylint: disable=too-few-public-methods
    """Pubsub exporter class."""

    def __init__(self):
        self.publisher = pubsub_v1.PublisherClient()

    def export(self, data, **config):
        """Export data to a Pub/Sub topic name.

        Args:
            data (dict): Data to send to Pub/Sub topic.
            config (dict): Pub/Sub configuration.
                project_id (str): Project id.
                topic_name (str): Topic name.

        Returns:
            str: Pub/Sub topic id.
        """
        project_id = config['project_id']
        topic_name = config['topic_name']
        topic_path = self.publisher.topic_path(project_id, topic_name)
        data = json.dumps(data, indent=4)
        data = data.encode('utf-8')
        return self.publisher.publish(topic_path, data=data).result()



# Copyright 2019 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#            http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
`bigquery.py`
BigQuery exporter class.
"""
import json
import logging
import google.api_core
from google.cloud import bigquery

LOGGER = logging.getLogger(__name__)


class BigqueryExporter:
    """BigQuery exporter class."""

    def __init__(self):
        self.client = bigquery.Client(project="unset")

    def export(self, data, **config):
        """Export results to BigQuery.

        Args:
            data (dict): Data to export.
                service_name (str): Service name.
                feature_name (str): Feature name.
                slo_name (str): SLO name.
                timestamp_human (str): Timestamp in human-readable format.
                measurement_window_seconds (int): Measurement window (in s).

            config (dict): Exporter config.
                project_id (str): BigQuery dataset project id.
                dataset_id (str): BigQuery dataset id.
                table_id (str): BigQuery table id.

        Raises:
            BigQueryError (object): BigQuery exception object.
        """
        project_id = config['project_id']
        dataset_id = config['dataset_id']
        table_id = config['table_id']
        self.client.project = project_id
        table_ref = self.client.dataset(dataset_id).table(table_id)
        try:
            table = self.client.get_table(table_ref)
        except google.api_core.exceptions.NotFound:
            table = self.create_table(project_id,
                                      dataset_id,
                                      table_id,
                                      schema=TABLE_SCHEMA)
        row_ids = "%s%s%s%s%s" % (data["service_name"], data["feature_name"],
                                  data["slo_name"], data["timestamp_human"],
                                  data["window"])
        results = self.client.insert_rows_json(
            table,
            json_rows=[data],
            row_ids=[row_ids],
            retry=google.api_core.retry.Retry(deadline=30))
        if results != []:
            raise BigQueryError(results)

        return results

    def create_table(self, project_id, dataset_id, table_id, schema=None):
        """Creates a BigQuery table from a schema.

        Args:
            project_id (str): Project id.
            dataset_id (str): Dataset id.
            table_id (str): Table id to create.
            schema (dict): BigQuery table schema in JSON format.
        """
        pyschema = []
        if schema is not None:
            schema = TABLE_SCHEMA
        for row in schema:
            field = bigquery.SchemaField(row['name'],
                                         row['type'],
                                         mode=row['mode'])
            pyschema.append(field)
        table_name = f"{project_id}.{dataset_id}.{table_id}"
        LOGGER.info(f"Creating table {table_name}", table_name)
        table = bigquery.Table(table_name, schema=pyschema)
        return self.client.create_table(table)


class BigQueryError(Exception):
    """Exception raised whenever a BigQuery error happened.

    Args:
        errors (list): List of errors.
    """

    def __init__(self, errors):
        super().__init__(BigQueryError._format(errors))
        self.errors = errors

    @staticmethod
    def _format(errors):
        err = []
        for error in errors:
            err.extend(error['errors'])
        return json.dumps(err)


TABLE_SCHEMA = [{
    "description": "",
    "name": "service_name",
    "type": "STRING",
    "mode": "REQUIRED"
}, {
    "description": "",
    "name": "feature_name",
    "type": "STRING",
    "mode": "REQUIRED"
}, {
    "description": "",
    "name": "slo_name",
    "type": "STRING",
    "mode": "REQUIRED"
}, {
    "description": "",
    "name": "slo_target",
    "type": "FLOAT64",
    "mode": "NULLABLE"
}, {
    "description": "",
    "name": "slo_description",
    "type": "STRING",
    "mode": "REQUIRED"
}, {
    "description": "",
    "name": "error_budget_policy_step_name",
    "type": "STRING",
    "mode": "NULLABLE"
}, {
    "description": "",
    "name": "error_budget_remaining_minutes",
    "type": "FLOAT64",
    "mode": ""
}, {
    "description": "",
    "name": "consequence_message",
    "type": "STRING",
    "mode": "NULLABLE"
}, {
    "description": "",
    "name": "error_budget_minutes",
    "type": "FLOAT64",
    "mode": "NULLABLE"
}, {
    "description": "",
    "name": "error_minutes",
    "type": "FLOAT64",
    "mode": "NULLABLE"
}, {
    "description": "",
    "name": "error_budget_target",
    "type": "FLOAT64",
    "mode": "NULLABLE"
}, {
    "description": "",
    "name": "timestamp_human",
    "type": "TIMESTAMP",
    "mode": "REQUIRED"
}, {
    "description": "",
    "name": "timestamp",
    "type": "FLOAT64",
    "mode": "NULLABLE"
}, {
    "description": "",
    "name": "cadence",
    "type": "STRING",
    "mode": "NULLABLE"
}, {
    "description": "",
    "name": "window",
    "type": "INT64",
    "mode": "REQUIRED"
}, {
    "description": "",
    "name": "bad_events_count",
    "type": "INT64",
    "mode": "NULLABLE"
}, {
    "description": "",
    "name": "good_events_count",
    "type": "INT64",
    "mode": "NULLABLE"
}, {
    "description": "",
    "name": "sli_measurement",
    "type": "FLOAT64",
    "mode": "NULLABLE"
}, {
    "description": "",
    "name": "gap",
    "type": "FLOAT64",
    "mode": "NULLABLE"
}, {
    "description": "",
    "name": "error_budget_measurement",
    "type": "FLOAT64",
    "mode": "NULLABLE"
}, {
    "description": "",
    "name": "error_budget_burn_rate",
    "type": "FLOAT64",
    "mode": "NULLABLE"
}, {
    "description": "",
    "name": "alerting_burn_rate_threshold",
    "type": "FLOAT64",
    "mode": "NULLABLE"
}, {
    "description": "",
    "name": "alert",
    "type": "BOOL",
    "mode": "NULLABLE"
}]



# Copyright 2019 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#            http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
`stackdriver_service_monitoring.py`
Stackdriver Service Monitoring exporter class.
"""
import difflib
import json
import logging
import os

import google.api_core.exceptions
from google.cloud.monitoring_v3 import ServiceMonitoringServiceClient
from google.protobuf.json_format import MessageToJson

from slo_generator.backends.stackdriver import StackdriverBackend
from slo_generator.utils import dict_snake_to_caml

LOGGER = logging.getLogger(__name__)

SID_GAE = 'gae:{project_id}_{module_id}'
SID_CLOUD_ENDPOINT = 'ist:{project_id}-{service}'
SID_CLUSTER_ISTIO = (
    'ist:{project_id}-zone-{location}-{cluster_name}-{service_namespace}-'
    '{service_name}')
SID_MESH_ISTIO = ('ist:{mesh_uid}-{service_namespace}-{service_name}')


class StackdriverServiceMonitoringBackend:
    """Stackdriver Service Monitoring backend class.

    Args:
        project_id (str): Stackdriver host project id.
        client (google.cloud.monitoring_v3.ServiceMonitoringServiceClient):
            Existing Service Monitoring API client. Initialize a new client if
            omitted.
    """

    def __init__(self, project_id, client=None):
        self.project_id = project_id
        self.client = client
        if client is None:
            self.client = ServiceMonitoringServiceClient()
        self.parent = self.client.project_path(project_id)
        self.workspace_path = f'workspaces/{project_id}'
        self.project_path = f'projects/{project_id}'

    def good_bad_ratio(self, timestamp, window, slo_config):
        """Good bad ratio method.

        Args:
            timestamp (int): UNIX timestamp.
            window (int): Window in seconds.
            slo_config (dict): SLO configuration.

        Returns:
            dict: SLO config.
        """
        return self.retrieve_slo(timestamp, window, slo_config)

    def distribution_cut(self, timestamp, window, slo_config):
        """Distribution cut method.

        Args:
            timestamp (int): UNIX timestamp.
            window (int): Window in seconds.
            slo_config (dict): SLO configuration.

        Returns:
            dict: SLO config.
        """
        return self.retrieve_slo(timestamp, window, slo_config)

    def basic(self, timestamp, window, slo_config):
        """Basic method (automatic SLOs for GAE / GKE (Istio) and Cloud
        Endpoints).

        Args:
            timestamp (int): UNIX timestamp.
            window (int): Window in seconds.
            slo_config (dict): SLO configuration.

        Returns:
            dict: SLO config.
        """
        return self.retrieve_slo(timestamp, window, slo_config)

    def window(self, timestamp, window, slo_config):
        """Window-based SLI method.

        Args:
            timestamp (int): UNIX timestamp.
            window (int): Window in seconds.
            slo_config (dict): SLO configuration.

        Returns:
            dict: SLO config.
        """
        return self.retrieve_slo(timestamp, window, slo_config)

    # pylint: disable=unused-argument
    def delete(self, timestamp, window, slo_config):
        """Delete method.

        Args:
            timestamp (int): UNIX timestamp.
            window (int): Window in seconds.
            slo_config (dict): SLO configuration.

        Returns:
            dict: SLO config.
        """
        return self.delete_slo(window, slo_config)

    def retrieve_slo(self, timestamp, window, slo_config):
        """Get SLI value from Stackdriver Monitoring API.

        Args:
            timestamp (int): UNIX timestamp.
            window (int): Window in seconds.
            slo_config (dict): SLO configuration.

        Returns:
            dict: SLO config.
        """
        # Get or create service
        service = self.get_service(slo_config)
        if service is None:
            service = self.create_service(slo_config)
        LOGGER.debug(service)

        # Get or create SLO
        slo = self.get_slo(window, slo_config)
        if not slo:
            slo = self.create_slo(window, slo_config)
        LOGGER.debug(service)

        # Now that we have our SLO, retrieve the TimeSeries from Stackdriver
        # Monitoring API for that particular SLO id.
        metric_filter = SSM.build_slo_id(window, slo_config, full=True)
        filter = f"select_slo_counts(\"{metric_filter}\")"

        # Query SLO timeseries
        stackdriver = StackdriverBackend(self.project_id)
        timeseries = stackdriver.query(timestamp,
                                       window,
                                       filter,
                                       aligner='ALIGN_SUM',
                                       reducer='REDUCE_SUM',
                                       group_by=['metric.labels.event_type'])
        timeseries = list(timeseries)
        good_event_count, bad_event_count = SSM.count(timeseries)
        return (good_event_count, bad_event_count)

    @staticmethod
    def count(timeseries):
        """Extract good_count, bad_count tuple from Stackdriver Monitoring API
        response.

        Args:
            timeseries (list): List of timeseries objects.

        Returns:
            tuple: A tuple (good_event_count, bad_event_count).
        """
        good_event_count, bad_event_count = 0, 0
        for timeserie in timeseries:
            event_type = timeserie.metric.labels['event_type']
            value = timeserie.points[0].value.double_value
            if event_type == 'bad':
                bad_event_count = value
            elif event_type == 'good':
                good_event_count = value
        return good_event_count, bad_event_count

    def create_service(self, slo_config):
        """Create Service object in Stackdriver Service Monitoring API.

        Args:
            slo_config (dict): SLO configuration.

        Returns:
            dict: Stackdriver Service Monitoring API response.
        """
        LOGGER.debug("Creating service ...")
        service_json = SSM.build_service(slo_config)
        service_id = SSM.build_service_id(slo_config)
        service = self.client.create_service(self.project_path,
                                             service_json,
                                             service_id=service_id)
        LOGGER.info(
            f'Service "{service_id}" created successfully in Stackdriver '
            f'Service Monitoring API.')
        return SSM.to_json(service)

    def get_service(self, slo_config):
        """Get Service object from Stackdriver Service Monitoring API.

        Args:
            slo_config (dict): SLO configuration.

        Returns:
            dict: Service config.
        """

        # Look for API services in workspace matching our config.
        service_id = SSM.build_service_id(slo_config)
        services = list(self.client.list_services(self.workspace_path))
        matches = [
            service for service in services
            if service.name.split("/")[-1] == service_id
        ]

        # If no match is found for our service name in the API, raise an
        # exception if the service should have been auto-added (method 'basic'),
        # else output a warning message.
        if not matches:
            msg = (f'Service "{service_id}" does not exist in '
                   f'workspace "{self.project_id}"')
            method = slo_config['backend']['method']
            if method == 'basic':
                sids = [service.name.split("/")[-1] for service in services]
                LOGGER.debug(
                    f'List of services in workspace {self.project_id}: {sids}')
                LOGGER.error(msg)
                raise Exception(msg)
            LOGGER.error(msg)
            return None

        # Match found in API, return it.
        service = matches[0]
        LOGGER.debug(f'Found matching service "{service.name}"')
        return SSM.to_json(service)

    @staticmethod
    def build_service(slo_config):
        """Build service JSON in Stackdriver Monitoring API from SLO
        configuration.

        Args:
            slo_config (dict): SLO configuration.

        Returns:
            dict: Service JSON in Stackdriver Monitoring API.
        """
        service_id = SSM.build_service_id(slo_config)
        display_name = slo_config.get('service_display_name', service_id)
        service = {'display_name': display_name, 'custom': {}}
        return service

    @staticmethod
    def build_service_id(slo_config, dest_project_id=None, full=False):
        """Build service id from SLO configuration.

        Args:
            slo_config (dict): SLO configuration.
            dest_project_id (str, optional): Project id for service if different
                than the workspace project id.
            full (bool): If True, return full service resource id including
                project path.

        Returns:
            str: Service id.
        """
        service_name = slo_config['service_name']
        feature_name = slo_config['feature_name']
        backend = slo_config['backend']
        project_id = backend['project_id']
        measurement = backend['measurement']
        app_engine = measurement.get('app_engine')
        cluster_istio = measurement.get('cluster_istio')
        mesh_istio = measurement.get('mesh_istio')
        cloud_endpoints = measurement.get('cloud_endpoints')

        # Use auto-generated ids for 'custom' SLOs, use system-generated ids
        # for all other types of SLOs.
        if app_engine:
            service_id = SID_GAE.format_map(app_engine)
            dest_project_id = app_engine['project_id']
        elif cluster_istio:
            service_id = SID_CLUSTER_ISTIO.format_map(cluster_istio)
            dest_project_id = cluster_istio['project_id']
        elif mesh_istio:
            service_id = SID_MESH_ISTIO.format_map(mesh_istio)
        elif cloud_endpoints:
            service_id = SID_CLOUD_ENDPOINT.format_map(cloud_endpoints)
            dest_project_id = cluster_istio['project_id']
        else:
            service_id = f'{service_name}-{feature_name}'

        if full:
            if dest_project_id:
                return f'projects/{dest_project_id}/services/{service_id}'
            return f'projects/{project_id}/services/{service_id}'

        return service_id

    def create_slo(self, window, slo_config):
        """Create SLO object in Stackdriver Service Monitoring API.

        Args:
            window (int): Window (in seconds).
            slo_config (dict): SLO config.

        Returns:
            dict: Service Management API response.
        """
        slo_json = SSM.build_slo(window, slo_config)
        slo_id = SSM.build_slo_id(window, slo_config)
        parent = SSM.build_service_id(slo_config, full=True)
        slo = self.client.create_service_level_objective(
            parent, slo_json, service_level_objective_id=slo_id)
        return SSM.to_json(slo)

    @staticmethod
    def build_slo(window, slo_config):  # pylint: disable=R0912,R0915
        """Get SLO JSON representation in Service Monitoring API from SLO
        configuration.

        Args:
            window (int): Window (in seconds).
            slo_config (dict): SLO Configuration.

        Returns:
            dict: SLO JSON configuration.
        """
        measurement = slo_config['backend'].get('measurement', {})
        method = slo_config['backend']['method']
        description = slo_config['slo_description']
        target = slo_config['slo_target']
        minutes, _ = divmod(window, 60)
        hours, _ = divmod(minutes, 60)
        display_name = f'{description} ({hours}h)'
        slo = {
            'display_name': display_name,
            'goal': target,
            'rolling_period': {
                'seconds': window
            }
        }
        filter_valid = measurement.get('filter_valid', "")
        if method == 'basic':
            methods = measurement.get('method', [])
            locations = measurement.get('location', [])
            versions = measurement.get('version', [])
            threshold = measurement.get('latency', {}).get('threshold')
            slo['service_level_indicator'] = {'basic_sli': {}}
            basic_sli = slo['service_level_indicator']['basic_sli']
            if methods:
                basic_sli['method'] = methods
            if locations:
                basic_sli['location'] = locations
            if versions:
                basic_sli['version'] = versions
            if threshold:
                basic_sli['latency'] = {
                    'threshold': {
                        'seconds': 0,
                        'nanos': int(threshold) * 10**6
                    }
                }
            else:
                basic_sli['availability'] = {}

        elif method == 'good_bad_ratio':
            filter_good = measurement.get('filter_good', "")
            filter_bad = measurement.get('filter_bad', "")
            slo['service_level_indicator'] = {
                'request_based': {
                    'good_total_ratio': {}
                }
            }
            sli = slo['service_level_indicator']
            ratio = sli['request_based']['good_total_ratio']
            if filter_good:
                ratio['good_service_filter'] = filter_good
            if filter_bad:
                ratio['bad_service_filter'] = filter_bad
            if filter_valid:
                ratio['total_service_filter'] = filter_valid

        elif method == 'distribution_cut':
            range_min = measurement.get('range_min', 0)
            range_max = measurement['range_max']
            slo['service_level_indicator'] = {
                'request_based': {
                    'distribution_cut': {
                        'distribution_filter': filter_valid,
                        'range': {
                            'max': float(range_max)
                        }
                    }
                }
            }
            sli = slo['service_level_indicator']['request_based']
            if range_min != 0:
                sli['distribution_cut']['range']['min'] = float(range_min)

        elif method == 'windows':
            filter = measurement.get('filter')
            # threshold = conf.get('threshold')
            # mean_in_range = conf.get('filter')
            # sum_in_range = conf.get('filter')
            slo['service_level_indicator'] = {
                'windows_based': {
                    'window_period': window,
                    'good_bad_metric_filter': filter,
                    # 'good_total_ratio_threshold': {
                    #   object (PerformanceThreshold)
                    # },
                    # 'metricMeanInRange': {
                    #   object (MetricRange)
                    # },
                    # 'metricSumInRange': {
                    #   object (MetricRange)
                    # }
                }
            }
        else:
            raise Exception(f'Method "{method}" is not supported.')
        return slo

    def get_slo(self, window, slo_config):
        """Get SLO object from Stackriver Service Monitoring API.

        Args:
            service_id (str): Service identifier.
            window (int): Window in seconds.
            slo_config (dict): SLO config.

        Returns:
            dict: API response.
        """
        service_path = SSM.build_service_id(slo_config, full=True)
        LOGGER.debug(f'Getting SLO for for "{service_path}" ...')
        slos = self.list_slos(service_path)
        slo_local_id = SSM.build_slo_id(window, slo_config)
        slo_json = SSM.build_slo(window, slo_config)
        slo_json = SSM.convert_slo_to_ssm_format(slo_json)

        # Loop through API response to find an existing SLO that corresponds to
        # our configuration.
        for slo in slos:
            slo_remote_id = slo['name'].split("/")[-1]
            equal = slo_remote_id == slo_local_id
            if equal:
                LOGGER.debug(f'Found existing SLO "{slo_remote_id}".')
                LOGGER.debug(f'SLO object: {slo}')
                strict_equal = SSM.compare_slo(slo_json, slo)
                if strict_equal:
                    return slo
                return self.update_slo(window, slo_config)
        LOGGER.warning('No SLO found matching configuration.')
        LOGGER.debug(f'SLOs from Stackdriver Monitoring API: {slos}')
        LOGGER.debug(f'SLO config converted: {slo_json}')
        return None

    def update_slo(self, window, slo_config):
        """Update an existing SLO.

        Args:
            window (int): Window (in seconds)
            slo_config (dict): SLO configuration.

        Returns:
            dict: API response.
        """
        slo_json = SSM.build_slo(window, slo_config)
        slo_id = SSM.build_slo_id(window, slo_config, full=True)
        LOGGER.warning(f"Updating SLO {slo_id} ...")
        slo_json['name'] = slo_id
        return SSM.to_json(self.client.update_service_level_objective(slo_json))

    def list_slos(self, service_path):
        """List all SLOs from Stackdriver Service Monitoring API.

        Args:
            service_path (str): Service path in the form
                'projects/{project_id}/services/{service_id}'.
            slo_config (dict): SLO configuration.

        Returns:
            dict: API response.
        """
        slos = self.client.list_service_level_objectives(service_path)
        slos = list(slos)
        LOGGER.debug(f"{len(slos)} SLOs found in Service Monitoring API.")
        # LOGGER.debug(slos)
        return [SSM.to_json(slo) for slo in slos]

    def delete_slo(self, window, slo_config):
        """Delete SLO from Stackdriver Monitoring API.

        Args:
            window (int): Window (in seconds).
            slo_config: SLO configuration.

        Returns:
            dict: API response.
        """
        slo_path = SSM.build_slo_id(window, slo_config, full=True)
        LOGGER.info(f'Deleting SLO "{slo_path}"')
        try:
            return self.client.delete_service_level_objective(slo_path)
        except google.api_core.exceptions.NotFound:
            LOGGER.warning(
                f'SLO "{slo_path}" does not exist in Service Monitoring API. '
                f'Skipping.')
            return None

    @staticmethod
    def build_slo_id(window, slo_config, full=False):
        """Build SLO id from SLO configuration.

        Args:
            slo_config (dict): SLO configuration.
            full (bool): If True, return full resource id including project.

        Returns:
            str: SLO id.
        """
        if 'slo_id' in slo_config:
            slo_id_part = slo_config['slo_id']
            slo_id = f'{slo_id_part}-{window}'
        else:
            slo_name = slo_config['slo_name']
            slo_id = f'{slo_name}-{window}'
        if full:
            service_path = SSM.build_service_id(slo_config, full=True)
            return f'{service_path}/serviceLevelObjectives/{slo_id}'
        return slo_id

    @staticmethod
    def compare_slo(slo1, slo2):
        """Compares 2 SLO configurations to see if they correspond to the same
        SLO.

        An SLO is deemed the same if the whole configuration is similar, except
        for the `goal` field that should be adjustable.

        Args:
            slo1 (dict): Service Monitoring API SLO configuration to compare.
            slo2 (dict): Service Monitoring API SLO configuration to compare.

        Returns:
            bool: True if the SLOs match, False otherwise.
        """
        exclude_keys = ["name"]
        slo1_copy = {k: v for k, v in slo1.items() if k not in exclude_keys}
        slo2_copy = {k: v for k, v in slo2.items() if k not in exclude_keys}
        local_json = json.dumps(slo1_copy, sort_keys=True)
        remote_json = json.dumps(slo2_copy, sort_keys=True)
        if os.environ.get('DEBUG') == '2':
            LOGGER.info("----------")
            LOGGER.info(local_json)
            LOGGER.info("----------")
            LOGGER.info(remote_json)
            LOGGER.info("----------")
            LOGGER.info(SSM.string_diff(local_json, remote_json))
        return local_json == remote_json

    @staticmethod
    def string_diff(string1, string2):
        """Diff 2 strings. Used to print comparison of JSONs for debugging.

        Args:
            string1 (str): String 1.
            string2 (str): String 2.

        Returns:
            list: List of messages pointing out differences.
        """
        lines = []
        for idx, string in enumerate(difflib.ndiff(string1, string2)):
            if string[0] == ' ':
                continue
            if string[0] == '-':
                info = u'Delete "{}" from position {}'.format(string[-1], idx)
                lines.append(info)
            elif string[0] == '+':
                info = u'Add "{}" to position {}'.format(string[-1], idx)
                lines.append(info)
        return lines

    @staticmethod
    def convert_slo_to_ssm_format(slo):
        """Convert SLO JSON to Service Monitoring API format.
        Address edge cases, like `duration` object computation.

        Args:
            slo (dict): SLO JSON object to be converted to Stackdriver Service
                Monitoring API format.

        Returns:
            dict: SLO configuration in Service Monitoring API format.
        """
        # Our local JSON is in snake case, convert it to Caml case.
        data = dict_snake_to_caml(slo)

        # The `rollingPeriod` field is in Duration format, convert it.
        try:
            period = data['rollingPeriod']
            data['rollingPeriod'] = SSM.convert_duration_to_string(period)
        except KeyError:
            pass

        # The `latency` field is in Duration format, convert it.
        try:
            latency = data['serviceLevelIndicator']['basicSli']['latency']
            threshold = latency['threshold']
            latency['threshold'] = SSM.convert_duration_to_string(threshold)
        except KeyError:
            pass

        return data

    @staticmethod
    def convert_duration_to_string(duration):
        """Convert a duration object to a duration string (in seconds).

        Args:
            duration (dict): Duration dictionary.

        Returns:
            str: Duration string.
        """
        duration_seconds = 0.000
        if 'seconds' in duration:
            duration_seconds += duration['seconds']
        if 'nanos' in duration:
            duration_seconds += duration['nanos'] * 10**(-9)
        if duration_seconds.is_integer():
            duration_str = int(duration_seconds)
        else:
            duration_str = "{:0.3f}".format(duration_seconds)
        return str(duration_str) + 's'

    @staticmethod
    def to_json(response):
        """Convert a Stackdriver Service Monitoring API response to JSON
        format.

        Args:
            response (obj): Response object.

        Returns:
            dict: Response object serialized as JSON.
        """
        return json.loads(MessageToJson(response))


SSM = StackdriverServiceMonitoringBackend



# Copyright 2019 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#            http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
`prometheus.py`
Prometheus backend implementation.
"""

import json
import logging
import os
import pprint
from prometheus_http_client import Prometheus

LOGGER = logging.getLogger(__name__)


class PrometheusBackend:
    """Backend for querying metrics from Prometheus."""

    def __init__(self, client=None, url=None, headers=None):
        self.client = client
        if not self.client:
            if url:
                os.environ['PROMETHEUS_URL'] = url
            if headers:
                os.environ['PROMETHEUS_HEAD'] = json.dumps(headers)
            LOGGER.debug(f'Prometheus URL: {url}')
            LOGGER.debug(f'Prometheus headers: {headers}')
            self.client = Prometheus()

    def query_sli(self, timestamp, window, slo_config):
        """Query SLI value from a given PromQL expression.

        Args:
            timestamp (int): UNIX timestamp.
            window (int): Window (in seconds).
            slo_config (dict): SLO configuration.

        Returns:
            float: SLI value.
        """
        conf = slo_config['backend']
        measurement = conf['measurement']
        expr = measurement['expression']
        expression = expr.replace("[window]", f"[{window}s]")
        data = self.query(expression, timestamp)
        LOGGER.debug(
            f"Expression: {expression} | Result: {pprint.pformat(data)}")
        try:
            sli_value = float(data['data']['result'][0]['value'][1])
        except IndexError:
            sli_value = 0
        LOGGER.debug(f"SLI value: {sli_value}")
        return sli_value

    def good_bad_ratio(self, timestamp, window, slo_config):
        """Compute good bad ratio from two metric filters.

        Args:
            timestamp (int): UNIX timestamp.
            window (int): Window (in seconds).
            slo_config (dict): SLO configuration.

        Note:
            At least one of `filter_bad` or `filter_valid` is required.

        Returns:
            tuple: A tuple of (good_count, bad_count).
        """
        conf = slo_config['backend']
        filter_good = conf['measurement']['filter_good']
        filter_bad = conf['measurement'].get('filter_bad')
        filter_valid = conf['measurement'].get('filter_valid')

        # Replace window by its value in the error budget policy step
        expr_good = filter_good.replace('[window]', f'[{window}s]')
        res_good = self.query(expr_good)
        good_count = PrometheusBackend.count(res_good)

        if filter_bad:
            expr_bad = filter_bad.replace('[window]', f'[{window}s]')
            res_bad = self.query(expr_bad, timestamp)
            bad_count = PrometheusBackend.count(res_bad)
        elif filter_valid:
            expr_valid = filter_valid.replace('[window]', f'[{window}s]')
            res_valid = self.query(expr_valid, timestamp)
            bad_count = PrometheusBackend.count(res_valid) - good_count
        else:
            raise Exception("`filter_bad` or `filter_valid` is required.")

        LOGGER.debug(f'Good events: {good_count} | ' f'Bad events: {bad_count}')

        return (good_count, bad_count)

    def query(self, filter, timestamp=None):  # pylint: disable=unused-argument
        """Query Prometheus server.

        Args:
            filter (str): Query filter.
            timestamp (int): UNIX timestamp.

        Returns:
            dict: Response.
        """
        response = self.client.query(metric=filter)
        response = json.loads(response)
        LOGGER.debug(pprint.pformat(response))
        return response

    @staticmethod
    def count(response):
        """Count events in Prometheus response.

        Args:
            response (dict): Prometheus query response.

        Returns:
            int: Event count.
        """
        # Note: this function could be replaced by using the `count_over_time`
        # function that Prometheus provides.
        try:
            return len(response['data']['result'][0]['values'])
        except (IndexError, KeyError) as exception:
            LOGGER.warning("Couldn't find any values in timeseries response")
            LOGGER.debug(exception)
            return 0  # no events in timeseries



# Copyright 2019 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#            http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
`report.py`
Report utilities.
"""

import logging
from dataclasses import dataclass, asdict, fields
from slo_generator import utils

LOGGER = logging.getLogger(__name__)


@dataclass(init=False)
class SLOReport:
    """SLO report dataclass. Compute an SLO report out of an SLO config and an
    Error Budget Policy step.

    Args:
        config (dict): SLO configuration.
        step (dict): Error budget policy step configuration.
        timestamp (int): Timestamp.
        client (obj): Existing backend client.
        delete (bool): Backend delete action.
    """
    # pylint: disable=too-many-instance-attributes

    # SLO
    service_name: str
    feature_name: str
    slo_name: str
    slo_target: float
    slo_description: str
    sli_measurement: float = 0
    bad_events_count: int = 0
    good_events_count: int = 0
    gap: float

    # Error budget
    error_budget_policy_step_name: str
    error_budget_target: float
    error_budget_measurement: float
    error_budget_burn_rate: float
    error_budget_minutes: float
    error_budget_remaining_minutes: float
    error_minutes: float

    # Error Budget step config
    timestamp: int
    timestamp_human: str
    window: int
    alert: bool
    alerting_burn_rate_threshold: float
    consequence_message: str

    def __init__(self, config, step, timestamp, client=None, delete=False):

        # Init dataclass fields from SLO config and Error Budget Policy
        self.__set_fields(**config,
                          **step,
                          lambdas={
                              'slo_target': float,
                              'alerting_burn_rate_threshold': int
                          })

        # Set other fields
        self.window = int(step['measurement_window_seconds'])
        self.timestamp = int(timestamp)
        self.timestamp_human = utils.get_human_time(timestamp)

        # Get backend results
        result = self.run_backend(config, client=client, delete=delete)
        if result:
            self.build(step, result)

    def build(self, step, result):
        """Compute all data necessary for the SLO report.

        Args:
            step (dict): Error Budget Policy step configuration.
            result (obj): Backend result.

        See https://landing.google.com/sre/workbook/chapters/implementing-slos/
        for details on the calculations.
        """
        info = self.__get_info()
        LOGGER.debug(f"{info} | SLO report starting ...")

        # SLI, Good count, Bad count, Gap from backend results
        sli, good_count, bad_count = self.get_sli(result)
        gap = sli - self.slo_target

        # Error Budget calculations
        eb_target = 1 - self.slo_target
        eb_value = 1 - sli
        eb_remaining_minutes = self.window * gap / 60
        eb_target_minutes = self.window * eb_target / 60
        eb_minutes = self.window * eb_value / 60
        if eb_target == 0:
            eb_burn_rate = 0
        else:
            eb_burn_rate = round(eb_value / eb_target, 1)

        # Alert boolean on burn rate excessive speed.
        alert = eb_burn_rate > self.alerting_burn_rate_threshold

        # Manage alerting message.
        if alert:
            consequence_message = step['overburned_consequence_message']
        elif eb_burn_rate <= 1:
            consequence_message = step['achieved_consequence_message']
        else:
            consequence_message = \
                'Missed for this measurement window, but not enough to alert'

        # Set fields in dataclass.
        self.__set_fields(sli_measurement=sli,
                          good_events_count=good_count,
                          bad_events_count=bad_count,
                          gap=gap,
                          error_budget_target=eb_target,
                          error_budget_measurement=eb_value,
                          error_budget_burn_rate=eb_burn_rate,
                          error_budget_remaining_minutes=eb_remaining_minutes,
                          error_budget_minutes=eb_target_minutes,
                          error_minutes=eb_minutes,
                          alert=alert,
                          consequence_message=consequence_message)

    def run_backend(self, config, client=None, delete=False):
        """Get appropriate backend method from SLO configuration and run it on
        current SLO config and Error Budget Policy step.

        Args:
            config (dict): SLO configuration.
            client (obj, optional): Backend client initiated beforehand.
            delete (bool, optional): Set to True if we're running a delete
                action.

        Returns:
            obj: Backend result.
        """
        info = self.__get_info()

        # Grab backend class and method dynamically.
        cfg = config.get('backend', {})
        cls = cfg.get('class')
        method = cfg.get('method')
        excluded_keys = ['class', 'method', 'measurement']
        backend_cfg = {k: v for k, v in cfg.items() if k not in excluded_keys}
        instance = utils.get_backend_cls(cls)(client=client, **backend_cfg)
        method = getattr(instance, method)
        LOGGER.debug(f'{info} | '
                     f'Using backend {cls}.{method.__name__} (from '
                     f'SLO config file).')

        # Delete mode activation.
        if delete and hasattr(instance, 'delete'):
            method = instance.delete
            LOGGER.warning(f'{info} | Delete mode enabled.')

        # Run backend method and return results.
        result = method(self.timestamp, self.window, config)
        LOGGER.debug(f'{info} | Backend results: {result}')
        return result

    def get_sli(self, result):
        """Compute SLI value and good / bad counts from the backend result.

        Some backends (e.g: Prometheus) are computing and returning the SLI
        value directly, others are sending a tuple (good_count, bad_count) and
        SLI value is computed from there.

        Args:
            result (obj): Backend result.

        Returns:
            tuple: A tuple of 3 values to unpack (float, int, int).
                float: SLI value.
                int: Good events count.
                int: Bad events count.

        Raises:
            Exception: When the backend does not return a proper result.
        """
        info = self.__get_info()
        if isinstance(result, tuple):
            good_count, bad_count = result
            LOGGER.debug(f"{info} | Good: {good_count} | Bad: {bad_count}")
            if (good_count + bad_count) == 0:
                LOGGER.error(f"{info} | No events found.")
                return 0, 0, 0
            sli_measurement = round(good_count / (good_count + bad_count), 6)

        elif isinstance(result, (float, int)):
            good_count, bad_count = 0, 0
            sli_measurement = round(result, 6)

        else:
            msg = "Backend did not return any valid results."
            LOGGER.error(msg)
            LOGGER.debug(f'Backend result: {result}')
            raise Exception(msg)

        return sli_measurement, good_count, bad_count

    def to_json(self):
        """Serialize dataclass to JSON."""
        return asdict(self)

    def is_empty(self):
        """Return True if report is empty (SLI = 0 or sum of good / bad count
        is null)."""
        return self.sli_measurement == 0

    def __set_fields(self, lambdas={}, **kwargs):
        """Set all fields in dataclasses from configs passed and apply function
        on values whose key match one in the dictionaries.

        Args:
            lambdas (dict): Dict {key: function} to apply a function on certain
            kwargs (dict): Dict of key / values to set in dataclass.
        """
        names = set(f.name for f in fields(self))
        for name in names:
            if name not in kwargs:
                continue
            value = kwargs[name]
            if name in lambdas.keys():
                value = lambdas[name](value)
            setattr(self, name, value)

    def __get_info(self):
        """Get info message describing current SLO andcurrent Error Budget Step.
        """
        slo_full_name = self.__get_slo_full_name()
        step_name = self.error_budget_policy_step_name
        return f"{slo_full_name :<32} | {step_name :<8}"

    def __get_slo_full_name(self):
        """Compile full SLO name from SLO configuration.

        Returns:
            str: Full SLO name.
        """
        return f'{self.service_name}/{self.feature_name}/{self.slo_name}'

    def __str__(self):
        report = self.to_json()
        info = self.__get_info()
        slo_target_per = self.slo_target * 100
        sli_per = round(self.sli_measurement * 100, 6)
        gap = round(self.gap * 100, 2)
        gap_str = str(gap)
        if gap >= 0:
            gap_str = f'+{gap}'
        sli_str = (f'SLI: {sli_per:<7} % | SLO: {slo_target_per} % | '
                   f'Gap: {gap_str:<6}%')
        result_str = (
            "BR: {error_budget_burn_rate:<2} / "
            "{alerting_burn_rate_threshold} | "
            "Alert: {alert} | Timestamp: {timestamp}").format_map(report)
        return f'{info} | {sli_str} | {result_str}'



# Copyright 2019 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#            http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
`stackdriver.py`
Stackdriver Monitoring exporter class.
"""
import logging
from google.cloud import monitoring_v3

LOGGER = logging.getLogger(__name__)
DEFAULT_METRIC_TYPE = "custom.googleapis.com/error_budget_burn_rate"
DEFAULT_METRIC_DESCRIPTION = ("Speed at which the error budget for a given"
                              "aggregation window is consumed")


class StackdriverExporter:
    """Stackdriver Monitoring exporter class."""

    def __init__(self):
        self.client = monitoring_v3.MetricServiceClient()

    def export(self, data, **config):
        """Export data to Stackdriver Monitoring.

        Args:
            data (dict): Data to send to Stackdriver Monitoring.
            config (dict): Stackdriver Monitoring metric config.
                project_id (str): Stackdriver host project id.
                custom_metric_type (str): Custom metric type.
                custom_metric_unit (str): Custom metric unit.

        Returns:
            object: Stackdriver Monitoring API result.
        """
        self.create_metric_descriptor(**config)
        self.create_timeseries(data, **config)

    def create_timeseries(self, data, **config):
        """Create Stackdriver Monitoring timeseries.

        Args:
            data (dict): Data to send to Stackdriver Monitoring.
            config (dict): Metric config.

        Returns:
            object: Metric descriptor.
        """
        series = monitoring_v3.types.TimeSeries()
        series.metric.type = config.get('metric_type', DEFAULT_METRIC_TYPE)

        # Write timeseries metric labels.
        series.metric.labels['error_budget_policy_step_name'] = str(
            data['error_budget_policy_step_name'])
        series.metric.labels['window'] = str(data['window'])
        series.metric.labels['service_name'] = data['service_name']
        series.metric.labels['feature_name'] = data['feature_name']
        series.metric.labels['slo_name'] = data['slo_name']
        series.metric.labels['alerting_burn_rate_threshold'] = str(
            data['alerting_burn_rate_threshold'])

        # Use the generic resource 'global'.
        series.resource.type = 'global'
        series.resource.labels['project_id'] = config['project_id']

        # Create a new data point.
        point = series.points.add()

        # Define end point timestamp.
        timestamp = data['timestamp']
        point.interval.end_time.seconds = int(timestamp)
        point.interval.end_time.nanos = int(
            (timestamp - point.interval.end_time.seconds) * 10**9)

        # Set the metric value.
        point.value.double_value = data['error_budget_burn_rate']

        # Record the timeseries to Stackdriver Monitoring.
        project = self.client.project_path(config['project_id'])
        result = self.client.create_time_series(project, [series])
        labels = series.metric.labels
        LOGGER.debug(
            f"timestamp: {timestamp} burnrate: {point.value.double_value}"
            f"{labels['service_name']}-{labels['feature_name']}-"
            f"{labels['slo_name']}-{labels['error_budget_policy_step_name']}")
        return result

    def create_metric_descriptor(self, **config):
        """Create Stackdriver Monitoring metric descriptor.

        Args:
            config (dict): Metric config.

        Returns:
            object: Metric descriptor.
        """
        project = self.client.project_path(config['project_id'])
        descriptor = monitoring_v3.types.MetricDescriptor()
        descriptor.type = config.get('metric_type', DEFAULT_METRIC_TYPE)
        descriptor.metric_kind = (
            monitoring_v3.enums.MetricDescriptor.MetricKind.GAUGE)
        descriptor.value_type = (
            monitoring_v3.enums.MetricDescriptor.ValueType.DOUBLE)
        descriptor.description = config.get('metric_description',
                                            DEFAULT_METRIC_DESCRIPTION)
        self.client.create_metric_descriptor(project, descriptor)
        return descriptor



# Copyright 2018 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#            http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A setuptools based setup module.

See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

from setuptools import setup, find_packages
from os import path
from io import open

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(name='slo-generator',
      version='1.0.0',
      description='SLO generator',
      long_description=long_description,
      long_description_content_type='text/markdown',
      author='Google Inc.',
      author_email='brunoreboul@google.com,ocervello@google.com',
      packages=find_packages(exclude=['contrib', 'docs', 'tests']),
      classifiers=[
          'Development Status :: 3 - Alpha',
          'Intended Audience :: Developers',
          'Topic :: Software Development :: Build Tools',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.4',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
      ],
      keywords='slo sli generator gcp',
      install_requires=[
          'google-api-python-client', 'oauth2client',
          'google-cloud-monitoring', 'google-cloud-pubsub',
          'google-cloud-bigquery', 'prometheus-http-client',
          'prometheus-client', 'pyyaml', 'opencensus', 'elasticsearch', 'pytz'
      ],
      entry_points={
          'console_scripts': [
              'slo-generator=slo_generator.cli:main',
          ],
      },
      python_requires='>=3.4')
